package com.company.surveypoc.screen.surveyresponse;

import com.company.surveypoc.entity.*;
import com.company.surveypoc.service.SurveyResponseService;
import io.jmix.core.EntityStates;
import io.jmix.core.Metadata;
import io.jmix.ui.AppUI;
import io.jmix.ui.UiComponents;
import io.jmix.ui.component.*;
import io.jmix.ui.model.CollectionPropertyContainer;
import io.jmix.ui.model.DataContext;
import io.jmix.ui.model.InstanceContainer;
import io.jmix.ui.screen.*;
import org.apache.commons.collections4.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;

import java.util.*;

@UiController("SurveyResponse.edit")
@UiDescriptor("survey-response-edit.xml")
@EditedEntityContainer("surveyResponseDc")
public class SurveyResponseEdit extends StandardEditor<SurveyResponse> {
    private static final Logger log = LoggerFactory.getLogger(SurveyResponseEdit.class);
    @Autowired
    private UiComponents uiComponents;
    @Autowired
    private SurveyResponseService surveyResponseService;
    @Autowired
    private VBoxLayout questionsBox;
    @Autowired
    private DataContext dataContext;
    @Autowired
    private Metadata metadata;
    @Autowired
    private HtmlAttributes htmlAttributes;
    @Autowired
    private CollectionPropertyContainer<SurveyAnswer> surveyAnswersDc;
    @Autowired
    private EntityComboBox<Survey> surveyField;
    @Autowired
    private EntityStates entityStates;
    private static final String COMPONENT_STYLE = "margin-left: 1em; margin-top: .5em; margin-bottom: 1em";
    private static final String COMPONENT_WIDTH = "90%";
    private static final Integer TEXTAREA_ROWS = 7;
    private static final String QUESTION_REQUIRED_MESSAGE = "'%s' is required";
    private List<SurveyQuestion> surveyQuestions;

    @Subscribe
    public void onBeforeShow(final BeforeShowEvent event) {
        surveyField.setEditable(entityStates.isNew(getEditedEntity()));
    }

    @Subscribe("surveyField")
    public void onSurveyFieldValueChange(final HasValue.ValueChangeEvent<Survey> event) {
        surveyQuestions = buildSurvey();
    }

    @Subscribe("hideBtn")
    public void onHideBtnClick(final Button.ClickEvent event) {
        for (SurveyQuestion question : surveyQuestions) {
            String questionComponentId = getQuestionComponentId(question);
            Component component = Objects.requireNonNull(questionsBox.getComponent(questionComponentId));
            if (component instanceof RadioButtonGroup<?>) {
                RadioButtonGroup<SurveyAnswerOption> option = (RadioButtonGroup<SurveyAnswerOption>) component;
                option.setRequired(false);
            }
            component.setVisible(false);
            log.info("Invisible");
        }
    }

    private void hideAllQuestions() {
        for (SurveyQuestion question : surveyQuestions) {
            String questionComponentId = getQuestionComponentId(question);
            Component component = Objects.requireNonNull(questionsBox.getComponent(questionComponentId));
            component.setVisible(false);
        }
    }

    private String getQuestionComponentId(SurveyQuestion surveyQuestion) {
        Integer questionId = surveyQuestion.getId();

        return "surveyQuestion" + questionId;
    }

    private List<SurveyQuestion> buildSurvey() {
        Survey survey = getEditedEntity().getSurvey();
        List<SurveyQuestion> questions = new ArrayList<>();

        if (survey != null) {
            //Sort the questions by displayOrder
            questions = survey.getQuestions().stream()
                    .sorted(Comparator.comparing(SurveyQuestion::getDisplayOrder))
                    .toList();
            for (SurveyQuestion surveyQuestion : questions) {
                switch (surveyQuestion.getAnswerType()) {
                    case SINGLE_ANSWER_MULTIPLE_CHOICE -> buildRadioButtonGroup(surveyQuestion);
                    case MULTIPLE_ANSWER_MULTIPLE_CHOICE -> buildCheckboxGroup(surveyQuestion);
                    case WRITTEN_SINGLE_LINE -> buildTextField(surveyQuestion);
                    case WRITTEN_MULTIPLE_LINE -> buildTextArea(surveyQuestion);
                }
            }
        }

        return questions;
    }

    private void buildRadioButtonGroup(SurveyQuestion surveyQuestion) {
        RadioButtonGroup<SurveyAnswerOption> radioButtonGroup = uiComponents.create(RadioButtonGroup.NAME);
        radioButtonGroup.setId(getQuestionComponentId(surveyQuestion));
        String caption = "<b>" + surveyQuestion.getQuestionText() + "</b>";
        radioButtonGroup.setCaption(caption);
        radioButtonGroup.setCaptionAsHtml(true);
        radioButtonGroup.setRequired(true);
        radioButtonGroup.setRequiredMessage(String.format(QUESTION_REQUIRED_MESSAGE, surveyQuestion.getQuestionText()));
        radioButtonGroup.setVisible(true);
        radioButtonGroup.setOptionsList(surveyResponseService.getAnswerOptions(surveyQuestion));
        radioButtonGroup.setWidth(COMPONENT_WIDTH);
        htmlAttributes.applyCss(radioButtonGroup, COMPONENT_STYLE);

        if (!surveyAnswersDc.getItems().isEmpty()) {
            Optional<SurveyAnswer> answerOptional = surveyAnswersDc.getItems().stream()
                    .filter(surveyAnswer -> surveyQuestion.equals(surveyAnswer.getSurveyQuestion()))
                    .findFirst();
            if (answerOptional.isPresent()) {
                SurveyAnswer surveyAnswer = answerOptional.get();
                SurveyAnswerOption surveyAnswerOption = surveyAnswer.getSurveyAnswerOption();
                radioButtonGroup.setValue(surveyAnswerOption);
            }
        }

        radioButtonGroup.addValueChangeListener(event -> {
            if (event.isUserOriginated()) {
                SurveyAnswer surveyAnswerToSave = surveyAnswersDc.getMutableItems().stream()
                        .filter(surveyAnswer -> surveyAnswer.getSurveyQuestion().equals(surveyQuestion))
                        .findFirst()
                        .orElseGet(() -> {
                            SurveyAnswer surveyAnswer = metadata.create(SurveyAnswer.class);
                            surveyAnswer.setSurveyResponse(getEditedEntity());
                            surveyAnswer.setSurveyQuestion(surveyQuestion);

                            return surveyAnswer;
                        });

                surveyAnswerToSave.setSurveyAnswerOption(event.getValue());
                SurveyAnswer mergedSurveyAnswer = dataContext.merge(surveyAnswerToSave);
                surveyAnswersDc.replaceItem(mergedSurveyAnswer);
            }
        });

        questionsBox.add(radioButtonGroup);
    }

    private void buildCheckboxGroup(SurveyQuestion surveyQuestion) {
        CheckBoxGroup<SurveyAnswerOption> checkBoxGroup = uiComponents.create(CheckBoxGroup.NAME);
        checkBoxGroup.setId(getQuestionComponentId(surveyQuestion));
        String caption = "<b>" + surveyQuestion.getQuestionText() + "</b>";
        checkBoxGroup.setCaption(caption);
        checkBoxGroup.setCaptionAsHtml(true);
        checkBoxGroup.setRequired(true);
        checkBoxGroup.setRequiredMessage(String.format(QUESTION_REQUIRED_MESSAGE, surveyQuestion.getQuestionText()));
        checkBoxGroup.setVisible(true);
        checkBoxGroup.setOptionsList(surveyResponseService.getAnswerOptions(surveyQuestion));
        checkBoxGroup.setWidth(COMPONENT_WIDTH);
        htmlAttributes.applyCss(checkBoxGroup, COMPONENT_STYLE);

        if (!surveyAnswersDc.getItems().isEmpty()) {
            List<SurveyAnswer> answers = surveyAnswersDc.getItems().stream()
                    .filter(surveyAnswer -> surveyQuestion.equals(surveyAnswer.getSurveyQuestion()))
                    .toList();

            if (!answers.isEmpty()) {
                checkBoxGroup.setValue(answers.stream().map(SurveyAnswer::getSurveyAnswerOption).toList());
            }
        }

        checkBoxGroup.addValueChangeListener(event -> {
            if (event.isUserOriginated()) {
                List<SurveyAnswer> surveyAnswers = surveyAnswersDc.getMutableItems().stream()
                        .filter(surveyAnswer -> surveyAnswer.getSurveyQuestion().equals(surveyQuestion))
                        .toList();

                surveyAnswersDc.getMutableItems().removeAll(surveyAnswers);
                // Stop tracking them
                surveyAnswers.forEach(dataContext::evict);

                if (CollectionUtils.isNotEmpty(event.getValue())) {
                    Collection<SurveyAnswerOption> answerOptions = event.getValue();

                    List<SurveyAnswer> answersToSave = new ArrayList<>(answerOptions.size());
                    for (SurveyAnswerOption answerOption : answerOptions) {
                        SurveyAnswer surveyAnswerToSave = surveyAnswers.stream()
                                .filter(answer -> answer.getSurveyAnswerOption().equals(answerOption))
                                .findFirst()
                                .orElseGet(() -> {
                                    SurveyAnswer surveyAnswer = metadata.create(SurveyAnswer.class);
                                    surveyAnswer.setSurveyResponse(getEditedEntity());
                                    surveyAnswer.setSurveyQuestion(surveyQuestion);
                                    return surveyAnswer;
                                });
                        surveyAnswerToSave.setSurveyAnswerOption(answerOption);
                        answersToSave.add(dataContext.merge(surveyAnswerToSave));
                    }
                    answersToSave.forEach(surveyAnswersDc::replaceItem);
                    List<SurveyAnswer> answersToRemove = new ArrayList<>(surveyAnswers);
                    answersToRemove.removeAll(answersToSave);
                    answersToRemove.forEach(dataContext::remove);
                }
            }
        });

        questionsBox.add(checkBoxGroup);
    }

    private void buildTextField(SurveyQuestion surveyQuestion) {
        TextField<String> textField = uiComponents.create(TextField.TYPE_STRING);
        textField.setId(getQuestionComponentId(surveyQuestion));
        String caption = "<b>" + surveyQuestion.getQuestionText() + "</b>";
        textField.setCaption(caption);
        textField.setCaptionAsHtml(true);
        textField.setRequired(true);
        textField.setRequiredMessage(String.format(QUESTION_REQUIRED_MESSAGE, surveyQuestion.getQuestionText()));
        textField.setVisible(true);
        textField.setWidth(COMPONENT_WIDTH);
        htmlAttributes.applyCss(textField, COMPONENT_STYLE);

        if (!surveyAnswersDc.getItems().isEmpty()) {
            Optional<SurveyAnswer> answerOptional = surveyAnswersDc.getItems().stream()
                    .filter(surveyAnswer -> surveyQuestion.equals(surveyAnswer.getSurveyQuestion()))
                    .findFirst();
            if (answerOptional.isPresent()) {
                SurveyAnswer surveyAnswer = answerOptional.get();
                String surveyAnswerFreeForm = surveyAnswer.getSurveyAnswerFreeForm();
                textField.setValue(surveyAnswerFreeForm);
            }
        }

        textField.addValueChangeListener(event -> {
            if (event.isUserOriginated()) {
                SurveyAnswer surveyAnswerToSave = surveyAnswersDc.getMutableItems().stream()
                        .filter(surveyAnswer -> surveyAnswer.getSurveyQuestion().equals(surveyQuestion))
                        .findFirst()
                        .orElseGet(() -> {
                            SurveyAnswer surveyAnswer = metadata.create(SurveyAnswer.class);
                            surveyAnswer.setSurveyResponse(getEditedEntity());
                            surveyAnswer.setSurveyQuestion(surveyQuestion);

                            return surveyAnswer;
                        });

                surveyAnswerToSave.setSurveyAnswerFreeForm(event.getValue());
                SurveyAnswer mergedSurveyAnswer = dataContext.merge(surveyAnswerToSave);
                surveyAnswersDc.replaceItem(mergedSurveyAnswer);
            }
        });

        questionsBox.add(textField);
    }

    private void buildTextArea(SurveyQuestion surveyQuestion) {
        TextArea<String> textArea = uiComponents.create(TextArea.TYPE_STRING);
        textArea.setId(getQuestionComponentId(surveyQuestion));
        String caption = "<b>" + surveyQuestion.getQuestionText() + "</b>";
        textArea.setCaption(caption);
        textArea.setCaptionAsHtml(true);
        textArea.setRequired(true);
        textArea.setRequiredMessage(String.format(QUESTION_REQUIRED_MESSAGE, surveyQuestion.getQuestionText()));
        textArea.setVisible(true);
        textArea.setWidth(COMPONENT_WIDTH);
        textArea.setRows(TEXTAREA_ROWS);
        htmlAttributes.applyCss(textArea, COMPONENT_STYLE);

        if (!surveyAnswersDc.getItems().isEmpty()) {
            Optional<SurveyAnswer> answerOptional = surveyAnswersDc.getItems().stream()
                    .filter(surveyAnswer -> surveyQuestion.equals(surveyAnswer.getSurveyQuestion()))
                    .findFirst();
            if (answerOptional.isPresent()) {
                SurveyAnswer surveyAnswer = answerOptional.get();
                String surveyAnswerFreeForm = surveyAnswer.getSurveyAnswerFreeForm();
                textArea.setValue(surveyAnswerFreeForm);
            }
        }

        textArea.addValueChangeListener(event -> {
            if (event.isUserOriginated()) {
                SurveyAnswer surveyAnswerToSave = surveyAnswersDc.getMutableItems().stream()
                        .filter(surveyAnswer -> surveyAnswer.getSurveyQuestion().equals(surveyQuestion))
                        .findFirst()
                        .orElseGet(() -> {
                            SurveyAnswer surveyAnswer = metadata.create(SurveyAnswer.class);
                            surveyAnswer.setSurveyResponse(getEditedEntity());
                            surveyAnswer.setSurveyQuestion(surveyQuestion);

                            return surveyAnswer;
                        });

                surveyAnswerToSave.setSurveyAnswerFreeForm(event.getValue());
                SurveyAnswer mergedSurveyAnswer = dataContext.merge(surveyAnswerToSave);
                surveyAnswersDc.replaceItem(mergedSurveyAnswer);
            }
        });

        questionsBox.add(textArea);
    }
}